.worker {
    font-family: "Courrier-New";
    border-width: 1px;
    width: 100%;
}
.worker td {
    border: 1px solid darkblue;
    box-sizing:content-box;
    padding-left: 10px;
}
.odd {background-color: lightblue;}
.worker-offline {background-color: red; 
    border-radius: 3px;
    padding:3px;
    color: white;}
.worker-paused {background-color: orange; 
    border-radius: 3px;
    padding:3px;
    color: white;}
.worker-running {background-color: green; 
    border-radius: 3px;
    padding:3px;
    color: white;}

.worker-concurrency button {
    padding:1px;
}

.total-div a {
    width: 15%;
    float:left;
    color: black;
    text-align:center;
    text-decoration: none;
    border: 1px solid;
    padding:10px;
    margin-bottom: 20px;
    background-image: -webkit-linear-gradient(top, #ffffff, #e6e6e6);
}

.awf-input  {
    width: 50px;
    margin: 0px 10px 0px 0px;
}

.awf-input-large  {
    width: 100px;
    margin: 0px 10px 0px 0px;
}

.awf-input-larger  {
    width: 150px;
    margin: 0px 10px 0px 0px;
}

.awf-input-largest  {
    width: 200px;
    margin: 0px 10px 0px 0px;
}


.awf-label {
    margin: 0px 0px 0px 10px;
}

.awf-button {
    margin: 2px;
}
