<!doctype html>
<html lang="en-US">

<head>
  <link rel="icon" type="image/png" href={{ url_for("static", filename="favicon.png") }}>
  <link href={{ url_for("static", filename="css/bootstrap.v5.0.2/bootstrap.min.css") }} rel="stylesheet">
  <script src={{ url_for("static", filename="js/bootstrap.v5.0.2/bootstrap.min.js") }}></script>
  <script src={{ url_for("static", filename="js/FileSaver.v2.0.4/FileSaver.min.js") }}></script>
  <script src={{ url_for("static", filename="js/jszip.min.v3.10.0.js") }}></script>
  <!-- <script src={{ url_for("static", filename="js/socket.io.v4.0.1/socket.io.js") }}></script> -->
  <script src="{{ url_for('static', filename='js/jquery-3.6.3.min.js') }}"></script>
  <script src="{{ url_for('static', filename='js/scitq-common.js') }}"></script>
  <script src="{{ url_for('static', filename='js/scitq-batch.js') }}"></script>
</head>

<body style="margin-top:-16px;">
    <title>batches view</title>
    <img src={{ url_for("static", filename="favicon.png") }} width=64px style="float:left; margin-bottom:2px;"/><h1 style="float:left;">batches view</h1>
    <p style='float:right; margin:0.5em;border-width: 1px;border-color: grey;background-color: lightgrey;padding:0.3em;max-height:2.5em;border-style: double;font-size:small'>v{{ package_version }} </p>
    <div class="container">
        <div class="">
      
        <p class="col-md-3" id="select"></p>
                  
                  
        </div> 
    </div>


    <div class="col-md-12"> 
      <div class="panel panel-default">
        <span id="batches"> 
          <table class="table table-responsive table-hover table-striped text-center ">
            <thead class="text-center table-secondary">
              <tr>
                <th>#</th>
                <th>Name</th>
                <th>Workers</th>
                <th>Progress</th>
                <th>pending</th>
                <th>running</th>
                <th>successes</th>
                <th>failed</th>
                <th>total</th>
                <th>Average duration [min - max] </th>
                <th>Actions</th>
              </tr>
            </thead>
            <tbody id="batch-table-body">
              <tr>
                <td>Loading please wait...</td>
              </tr>
            </tbody>
          </table>
        </span>
      </div>
    </div>

    <div class="clearfix"></div>
    <br /><br />
</body>

</html>


