[Unit]
Description=scitq-worker
After=multi-user.target

[Service]
Environment=PATH=/usr/bin:/usr/local/bin
EnvironmentFile=/etc/scitq-worker.conf
Type=simple
ExecStart=scitq-worker -s {{ status }} {{ target }} {{ concurrency }}

[Install]
WantedBy=multi-user.target